/*
Copyright (C) 2013, 2014 Bengt Martensson.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see http://www.gnu.org/licenses/.
*/

package org.harctoolbox.harchardware.ir;

import gnu.io.NoSuchPortException;
import gnu.io.PortInUseException;
import gnu.io.UnsupportedCommOperationException;
import java.io.IOException;
import java.util.ArrayList;
import org.harctoolbox.IrpMaster.IncompatibleArgumentException;
import org.harctoolbox.IrpMaster.IrSignal;
import org.harctoolbox.IrpMaster.IrpMasterException;
import org.harctoolbox.IrpMaster.IrpUtils;
import org.harctoolbox.IrpMaster.ModulatedIrSequence;
import org.harctoolbox.IrpMaster.Pronto;
import org.harctoolbox.harchardware.HarcHardwareException;
import org.harctoolbox.harchardware.comm.LocalSerialPort;
import org.harctoolbox.harchardware.comm.LocalSerialPortRaw;

/**
 * This class ...
 * @see <a href="http://www.dangerousprototypes.com/docs/USB_IR_Toy:_Sampling_mode">http://www.dangerousprototypes.com/docs/USB_IR_Toy:_Sampling_mode</a>
 *
 */
public final class IrToy extends IrSerial<LocalSerialPortRaw> implements IRawIrSender, ICapture {

    public static final String defaultPortName = "/dev/ttyACM0";
    public static final int defaultBaudRate = 9600;//115200;//9600;
    public static final LocalSerialPort.FlowControl defaultFlowControl = LocalSerialPort.FlowControl.RTSCTS;
    public static final int defaultTimeout = 2000;

    /*private String portName;
    private int baudRate;
    private LocalSerialPort.FlowControl flowControl;*/
    private static final int dataSize = 8;
    private static final int stopBits = 1;
    //private double fallbackFrequency = 38000;
    private static final LocalSerialPort.Parity parity = LocalSerialPort.Parity.NONE;

    private static final int bufsize = 255;

    private static final double oscillatorFrequency = 48000000;
    private static final double period = 21.3333; // microseconds
    private static final double PICClockFrequency = 12000000;
    private static final byte dutyCycle = 0; // semantically: don't care

    private static final boolean transmitNotifyEnabled = true;
    private static final boolean transmitByteCountReportEnabled = true;
    private static final boolean transmitHandshakeEnabled = true;

    private final static byte cmdReset = 0x00; // Reset (returns to remote decoder mode)
    // 0x01 RESERVED for SUMP RUN
    // 0x02 RESERVED for SUMP ID
    private final static byte cmdTransmit = 0x03; // Transmit (FW v07+)
    private final static byte cmdFrequencyReport = 0x04; // Frequency report (reserved for future hardware)
    // 0x05 Setup sample timer (FW v07+)
    private final static byte cmdSetFrequency = 0x06; // Setup frequency modulation timer (FW v07+)
    private final static byte cmdLedMuteOn = 0x10; // LED mute on (FW v07+)
    private final static byte cmdLedMuteOff = 0x11; // LED mute off (FW v07+)
    private final static byte cmdLedOn = 0x12; // LED on (FW v07+)
    private final static byte cmdLedOff = 0x13; // LED off (FW v07+)
    // 0x23 Settings descriptor report (FW v20+)
    private final static byte cmdTransmitByteCountReport = 0x24; // Enable transmit byte count report (FW v20+)
    private final static byte cmdTransmitNotify = 0x25; // Enable transmit notify on complete (FW v20+)
    private final static byte cmdTransmitHandshake = 0x26; // Enable transmit handshake (FW v20+)
    // 0x30 IO write
    // 0x31 IO direction
    // 0x32 IO read
    // 0x40 UART setup
    // 0x41 UART close
    // 0x42 UART write

    private final static byte cmdSamplingMode = (byte) 's';
    private final static byte cmdSelfTest = (byte) 't';
    private final static byte cmdVersion = (byte) 'v';
    private final static byte cmdBootloaderMode = (byte) '$';
    private final static byte endOfData = (byte) 0xff;
    private final static int transmitByteCountToken = (int) 't';
    private final static int transmitCompleteSuccess = (int) 'C';
    private final static int transmitCompleteFailure = (int) 'F';

    private boolean stopCaptureRequest = true;

    private String protocolVersion;
    private String version;

    //private int beginTimeout;
    //private int middleTimeout;
    //private int endingTimeout;
    //private final static int defaultBeginTimeout = 5000;
    //private final static int defaultMiddleTimeout = 1000;
    //private final static int defaultEndingTimeout = 500;

    IrToy() throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        this(defaultPortName, defaultBaudRate, defaultFlowControl, defaultTimeout, false);
    }

    public IrToy(String portName) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        this(portName, defaultBaudRate, defaultFlowControl, defaultTimeout, false);
    }

    public IrToy(String portName, int timeout, boolean verbose) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        this(portName, defaultBaudRate, defaultFlowControl, timeout, verbose);
    }

    public IrToy(String portName, int baudRate, int timeout, boolean verbose) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        this(portName, baudRate, defaultFlowControl, timeout, verbose);
    }

    public IrToy(String portName, int beginTimeout, int middleTimeout, int endingTimeout, boolean verbose) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        this(portName, defaultBaudRate, defaultFlowControl, beginTimeout, verbose);
    }

    public IrToy(String portName, int baud, int beginTimeout, int middleTimeout, int endingTimeout, boolean verbose) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        this(portName, baud, defaultFlowControl, beginTimeout, verbose);
    }

    public IrToy(String portName, int baudRate, LocalSerialPort.FlowControl flowControl, int timeout, boolean verbose) throws NoSuchPortException, PortInUseException, UnsupportedCommOperationException, IOException {
        super(LocalSerialPortRaw.class, portName, baudRate, dataSize, stopBits, parity, flowControl, timeout, verbose);
    }

    @Override
    public void open() throws IOException, HarcHardwareException {
        super.open();
        reset();
        send(cmdVersion);
        version = readString();
        send(cmdSamplingMode);
        try {
            Thread.sleep(100);
        } catch (InterruptedException ex) {
        }
        protocolVersion = readString();
        try {
            Thread.sleep(100);
        } catch (InterruptedException ex) {
        }
        if (transmitNotifyEnabled)
            send(cmdTransmitNotify);
        //Thread.sleep(100);
        if (transmitHandshakeEnabled)
            send(cmdTransmitHandshake);
        //Thread.sleep(100);
        if (transmitByteCountReportEnabled)
            send(cmdTransmitByteCountReport);
        try {
            Thread.sleep(100);
        } catch (InterruptedException ex) {
        }
    }

    public void reset() throws IOException {
        send(new byte[]{cmdReset, cmdReset, cmdReset, cmdReset, cmdReset});
        try {
            Thread.sleep(100);
        } catch (InterruptedException ex) {
        }
    }

    private void send(byte[] buf) throws IOException {
        serialPort.sendBytes(buf);
        serialPort.flush();
    }

    private void send(byte[] buf, int offset, int length) throws IOException {
        serialPort.sendBytes(buf, offset, length);
        serialPort.flush();
    }

    private void send(byte b) throws IOException {
        serialPort.sendByte(b);
        serialPort.flush();
    }

    private byte[] toByteArray(int[] data) {
        byte[] buf = new byte[2*data.length + 2];
        for (int i = 0; i < data.length; i++) {
            int periods = (int)Math.round(((double)data[i])/period);
            buf[2*i] = (byte)(periods / 256);
            buf[2*i+1] = (byte) (periods % 256);
        }
        buf[2*data.length] = endOfData;
        buf[2*data.length+1] = endOfData;
        return buf;
    }

    private int[] receive() throws IOException {
        ArrayList<Integer> array = new ArrayList<Integer>();
        stopCaptureRequest = false;
        while (! stopCaptureRequest) {
            int val = read2Bytes();
            int ms = (int) Math.round(val*period);
            array.add(ms);
            if (val == 0xffff)
                // Only way for timeout, 1.4 seconds. Too long for most use cases ... :-\
                break;
        }
        int[] result = new int[array.size()];
        for (int i = 0; i < array.size(); i++)
            result[i] = array.get(i);
        return result;
    }

    private double getFrequency(int onTimes) throws IOException {
        send(cmdFrequencyReport);
        /*int t1 =*/ read2Bytes();
        /*int t2 =*/ read2Bytes();
        /*int t3 =*/ read2Bytes();
        int count = read2Bytes();
        //System.err.println(t1);System.err.println(t2);System.err.println(t3);System.err.println(count);System.err.println(onTimes);
        //return (2*PICClockFrequency)/((double)(t3 - t1));
        return ((double)count)/(((double) onTimes) * IrpUtils.microseconds2seconds) ;
    }

    @Override
    public ModulatedIrSequence capture() throws IOException {
        int [] data = receive();
        if (! stopCaptureRequest) {
            int sum = 0;
            for (int i = 0; i < data.length/2; i++)
                sum += data[2*i];
            double frequency = getFrequency(sum);
            try {
                return new ModulatedIrSequence(data, frequency);
            } catch (IncompatibleArgumentException ex) {
                System.err.println("Internal error");
            }
        }
        return null;
    }

    @Override
    public boolean stopCapture() {
        stopCaptureRequest = true;
        return true;
    }

    private boolean transmit(int[] data, double frequency) throws IOException {
        if (frequency > 0)
            setFrequency(frequency);
        return transmit(data);
    }

    private boolean transmit(int[] data) throws IOException {
        send(cmdTransmit);
        byte[] buf = toByteArray(data);

        if (transmitHandshakeEnabled) {
            int bytesSent = 0;
            while (bytesSent < buf.length) {
                int noBytes = readByte();
                int toSend = Math.min(noBytes, buf.length - bytesSent);
                send(buf, bytesSent, toSend);
                bytesSent += toSend;
            }

        }
        readByte();

        if (transmitByteCountReportEnabled) {
            int token = readByte();
            if (token != transmitByteCountToken)
                return false;
            int bytesSent = read2Bytes();
            if (bytesSent != data.length*2+2)
                return false;
        }

        if (transmitNotifyEnabled) {
            int token = readByte();
            if (token != transmitCompleteSuccess)
                return false;
        }
        return true;
    }

    private String readString() throws IOException {
        byte[] buf = new byte[bufsize];
        int length = this.serialPort.readBytes(buf);
        return new String(buf, 0, length, IrpUtils.dumbCharset);
    }

    private int readByte() throws IOException {
        return serialPort.readByte();
    }

    private int read2Bytes() throws IOException {
        int high = readByte();
        int low = readByte();
        return 256*high + low;
    }

    public String selftest() throws IOException {
        reset();
        send(cmdSelfTest);
        byte[] buf = new byte[4];
        for (int i = 0; i < 4; i++)
            buf[i] = (byte) readByte();

        return new String(buf, IrpUtils.dumbCharset);
    }

    public void bootloaderMode() throws IOException {
        reset();
        send(cmdBootloaderMode);
    }

    public String getProtocolVersion() {
        return protocolVersion;
    }

    public void setLedMute(boolean status) throws IOException {
        send(status ? cmdLedMuteOn : cmdLedMuteOff);
    }

    public void setLed(boolean status) throws IOException {
        send(status ? cmdLedOn : cmdLedOff);
    }

    private void setFrequency(double frequency) throws IOException {
        byte pr2 = (byte) Math.round(oscillatorFrequency/(16*frequency) - 1);
        byte[] buf = new byte[3];
        buf[0] = cmdSetFrequency;
        buf[1] = pr2;
        buf[2] = dutyCycle;
        send(buf);
    }

    @Override
    public String getVersion() {
        return version;
    }

    @Override
    public void setTimeout(int beginTimeout, int maxLearnLength, int endTimeout) throws IOException {
        setTimeout(beginTimeout);
    }

    @Override
    public boolean sendIr(IrSignal code, int count, Transmitter transmitter) throws IrpMasterException, IOException {
        return transmit(code.toIntArray(count-1), code.getFrequency());
    }

    public boolean sendCcf(String ccf, int count, Transmitter transmitter) throws IOException, IrpMasterException {
        return sendIr(Pronto.ccfSignal(ccf), count, transmitter);
    }

    /**
     * Not supported due to hardware restrictions.
     *
     * @param ccf
     * @param transmitter
     * @return
     */
    public boolean sendCcfRepeat(String ccf, Transmitter transmitter) {
        throw new UnsupportedOperationException("Not supported due to hardware restrictions.");
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        IrToy toy = null;
        String portName = "/dev/ttyACM0";
        try {

            //int[] data = new int[]{889, 889, 1778, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 889, 90886};
            toy = new IrToy(portName);
            toy.open();
            System.out.println(toy.getVersion());
            //String result = toy.selftest();
            //System.out.println(result);
            toy.setLed(true);
            toy.setLedMute(false);
            //boolean success = toy.transmit(36000, data);
            IrSignal signal = new IrSignal("../IrpMaster/data/IrpProtocols.ini", "nec1", "D=122 F=26");
            boolean success = toy.sendIr(signal, 10, null);
            System.out.println(success);
            toy.close();
            //w.listen();
        } catch (NoSuchPortException ex) {
            System.err.println("Port for IRToy " + portName + " was not found");
        } catch (HarcHardwareException ex) {
            System.err.println(ex.getMessage());
        } catch (PortInUseException ex) {
            System.err.println("Port for IRToy in use");
        } catch (UnsupportedCommOperationException ex) {
            System.err.println("xxx" + ex.getMessage());
            System.err.println(ex.getMessage());
        } catch (IOException ex) {
            System.err.println("xxx" + ex.getMessage());
            //ex.printStackTrace();
        } catch (IrpMasterException ex) {
            System.err.println("xxx" + ex.getMessage());
            //ex.printStackTrace();
        } finally {
            if (toy != null)
                try {
                    toy.close();
                } catch (IOException ex) {
                }
        }
    }
}
